# from contextlib import redirect_stdout, redirect_stderr
# from io import StringIO
from unittest.mock import patch, call

from ..base import IntegrationTests
from ..util import fork


class TestNoProfile(IntegrationTests):
    """ Integration tests for no profile. """
    @fork()
    def test_logout_via_awscli(self):
        """ Logging out should throw AlreadyLoggedOut when not logged in. """
        mesg = 'Already logged out!\n'
        self.assertAwsCliReturns('logout', stderr=mesg, code=3)
        # TODO FIXME NO HARDCODE

    @fork()
    @patch('builtins.input', return_value='')
    def test_save_credentials_default_profile(self, mock_input):
        """ Creates a default entry in ~/.aws/credentials """
        calls = [
            call('ECP Endpoint URL [None]: '),
            call('Username [None]: '),
            call('Enable Keyring [False]: '),
            call('Duo Factor [None]: '),
            call('Role ARN [None]: '),
        ]

        self.assertAwsCliReturns('login', 'configure', code=0)
        mock_input.assert_has_calls(calls)
        self.assertEqual(len(mock_input.call_args_list), len(calls))


class TestDefaultProfile(IntegrationTests):
    """ Integration tests for default profile. """

    @fork()
    def test_logout_via_awscli(self):
        """ Logging out should throw AlreadyLoggedOut when not logged in. """
        self.profile = 'default'
        self.aws_credentials = """
[default]
aws_access_key_id = abc
aws_secret_access_key = def
aws_session_token = ghi
"""

        mesg = 'Already logged out!\n'
        self.assertAwsCliReturns('logout', stderr=mesg, code=3)
        # TODO FIXME NO HARDCODE ERROR CODES!



""" A collection of custom exceptions """
from .const import FACTORS


class AWSCLILogin(Exception):
    code = 1  # type: int


class ConfigError(AWSCLILogin):
    pass


class AlreadyLoggedIn(ConfigError):
    code = 2

    def __init__(self) -> None:
        mesg = 'Already logged in!'
        super().__init__(mesg)


class AlreadyLoggedOut(ConfigError):
    code = 3

    def __init__(self) -> None:
        mesg = 'Already logged out!'
        super().__init__(mesg)


class ProfileNotFound(ConfigError):
    code = 4

    def __init__(self, profile: str) -> None:
        mesg = 'The login profile (%s) could not be found!'
        super().__init__(mesg % profile)


class ProfileMissingArgs(ConfigError):
    code = 5

    def __init__(self, profile: str, *args: str) -> None:
        mesg = 'The login profile (%s) is missing argument(s): %s!'
        super().__init__(mesg % (profile, ', '.join(args)))


class InvalidFactor(ConfigError):
    code = 6

    def __init__(self, factor: str) -> None:
        valid_factors = ', '.join(FACTORS) + '.'
        mesg = "Invalid factor %s! Valid values are: " + valid_factors
        super().__init__(mesg % factor)


class SAML(AWSCLILogin):
    pass


class AuthnFailed(SAML):
    code = 7

    def __init__(self) -> None:
        mesg = 'Authentication failed!'
        super().__init__(mesg)


class InvalidSOAP(SAML):
    code = 8

    def __init__(self, url: str) -> None:
        mesg = "Invalid SOAP returned by ECP Endpoint: %s!\n" % url
        mesg += "ECP Endpoint URL may be bad!"
        super().__init__(mesg)


class MissingCookieJar(SAML):
    code = 9

    def __init__(self, cookies: str) -> None:
        mesg = "Failed to load cookie jar: %s"
        super().__init__(mesg % cookies)


class RoleParseFail(SAML):
    code = 10

    def __init__(self, role: str) -> None:
        mesg = "Bad SAML Response! Failed to parse role: %s!"
        super().__init__(mesg % role)



""" Base classes used for testing """
from typing import Any
from unittest.mock import patch

from awscli_login.config import Profile

from .util import MockSession
from ..util import login_cli_args
from ..base import CleanAWSLoginEnvironment


class ProfileBase(CleanAWSLoginEnvironment):
    """
    Base class for testing awscli_login.config.Profile.

    NOTA BENE: This class does NOT load self.profile and therefore is useful
    for testing failure conditions for Profile().
    """

    def Profile(self, profile: str = 'default', no_args: bool = False,
                **kwargs):
        args = None if no_args else login_cli_args(**kwargs)
        session = MockSession(profile)

        self.profile = Profile(session, args)
        return self.profile

    def assertProfileHasAttr(self, attr: str, evalue: Any) -> None:
        """ If called tests that profile has the given attribute
            and value. Returns None on success, and an error
            message on failure. """
        self.assertHasAttr(self.profile, attr, evalue)

    def assertProfileHasAttrs(self, **kwargs: Any) -> None:
        """ If called tests that profile has the expected
            attributes and values specifed in expected_attr_vals. """
        self.assertHasAttrs(self.profile, **kwargs)

    def patcher(self, target: Any, **kwargs):  # TODO FIXME
        patcher = patch(target, **kwargs)
        self.addCleanup(patcher.stop)

        return patcher.start()



from botocore.session import Session

from .__main__ import error_handler
from .config import Profile


@error_handler()
def configure(profile: Profile, session: Session):
    """ Interactive login profile configuration. """
    profile.update()



# from signal import signal, SIGINT, SIGTERM
import logging
import traceback

from argparse import Namespace
from datetime import datetime
from functools import wraps

import boto3

from botocore.session import Session
from daemoniker import Daemonizer, SignalHandler1
from daemoniker import send, SIGINT, SIGTERM, SIGABRT

from .config import (
    Profile,
    ERROR_NONE,
    ERROR_UNKNOWN,
)
from .exceptions import AlreadyLoggedOut, AWSCLILogin
from .logger import (
    configConsoleLogger,
    configFileLogger,
)
from .saml import (
    authenticate,
    refresh,
)
from .util import (
    get_selection,
    nap,
    remove_credentials,
    save_credentials,
    write_identity_file,
    remove_identity_file
)
from .typing import Role

logger = logging.getLogger(__package__)


def save_sts_token(session: Session, client, saml: str,
                   role: Role, duration) -> datetime:
    params = dict(
        RoleArn=role[1],
        PrincipalArn=role[0],
        SAMLAssertion=saml,
    )
    if duration:
        params['DurationSeconds'] = duration
        # duration is optional and can be set by the role;
        # avoid passing if not set.

    token = client.assume_role_with_saml(**params)
    logger.info("Retrieved temporary Amazon credentials for role: " + role[1])
    
    return save_credentials(session, token)


def daemonize(profile: Profile, session: Session, client: boto3.client,
              role: Role, expires: datetime) -> bool:
    with Daemonizer() as (is_setup, daemonizer):
        is_parent, profile, session, client, role, expires = daemonizer(
            profile.pidfile,
            profile,
            session,
            client,
            role,
            expires,
        )

        if not is_parent:
            sighandler = SignalHandler1(profile.pidfile)
            sighandler.start()

            logger = configFileLogger(profile.logfile, logging.INFO)
            logger.info('Startig refresh process for role %s' % role[1])

            # TODO add retries!
            while(True):
                retries = 0
                nap(expires, 0.9)

                while(True):
                    try:
                        saml, _ = refresh(
                            profile.ecp_endpoint_url,
                            profile.cookies,
                        )
                    except Exception as e:
                        retries += 1

                        if (retries < 4):
                            logger.info('Refresh failed: %s' % str(e))
                            nap(expires, 0.2)
                        else:
                            raise
                    else:
                        break
                
                expires = save_sts_token(session, client, saml, role, None)

        return is_parent


def error_handler(skip_args=True, validate=False):
    """ A decorator for exception handling and logging. """
    def decorator(f):
        @wraps(f)
        def wrapper(args: Namespace, session: Session):
            exp = None  # type: Exception
            code = ERROR_NONE
            sig = None

            try:
                configConsoleLogger(args.verbose)
                if not skip_args:
                    profile = Profile(session, args, validate)
                else:
                    profile = Profile(session, None, validate)

                f(profile, session)
            except AWSCLILogin as e:
                code = e.code
                exp = e
            except SIGINT:
                sig = 'SIGINT'
            except SIGABRT:
                sig = 'SIGABRT'
            except SIGTERM:
                sig = 'SIGTERM'
            except Exception as e:
                code = ERROR_UNKNOWN
                exp = e
            finally:
                if code:
                    logger.error(str(exp), exc_info=False)
                    logger.debug(traceback.format_exc())

                if sig:
                    logger.info('Received signal: %s. Shutting down...' % sig)

                exit(code)

        return wrapper
    return decorator


@error_handler(skip_args=False, validate=True)
def main(profile: Profile, session: Session):
    is_parent = True

    try:
        client = boto3.client('sts')

        # TODO force-refresh should kill refresh!
        if not profile.force_refresh:
            profile.raise_if_logged_in()

        # Must know username to lookup cookies
        profile.get_username()

        try:
            saml, roles = refresh(
                profile.ecp_endpoint_url,
                profile.cookies,
            )
        except Exception:
            creds = profile.get_credentials()
            saml, roles = authenticate(profile.ecp_endpoint_url,
                                       profile.cookies, *creds)

        duration = profile.duration
        role = get_selection(roles, profile.role_arn)
        expires = save_sts_token(session, client, saml, role, duration)
        _,role_arn = role
        account_id = role_arn.split(':')[4]
        write_identity_file(account_id)

        if not profile.force_refresh and not profile.disable_refresh:
            is_parent = daemonize(profile, session, client, role, expires)
    except Exception:
        raise
    finally:
        if not is_parent:
            logger.info('Exiting refresh process')


@error_handler()
def logout(profile: Profile, session: Session):
    try:
        send(profile.pidfile, SIGINT)
        remove_credentials(session)
    except IOError:
        raise AlreadyLoggedOut
    finally:
        remove_identity_file()
        



from copy import copy
from os import getpid, path
from os.path import isfile
from typing import Any, Dict  # NOQA
from unittest import SkipTest
from unittest.mock import patch

from awscli_login.config import JAR_DIR
from awscli_login.exceptions import (
    AlreadyLoggedIn,
    ProfileNotFound,
    ProfileMissingArgs,
)

from ..base import (
    TempDir,
)

from .base import ProfileBase
from .util import qremove


class NoProfile(ProfileBase):
    """ Given an empty configuration. """

    def test_empty_config(self) -> None:
        """ Loading empty profile should throw Exception ProfileNotFound. """
        self.login_config = ''

        with self.assertRaises(ProfileNotFound):
            self.Profile()


class MissingProfile(ProfileBase):
    """ Given the desired profile is missing """

    def test_missing_profile(self) -> None:
        """ Loading missing profile should throw Exception ProfileNotFound. """
        self.login_config = """
[default]
ecp_endpoint_url = foo
    """
        with self.assertRaises(ProfileNotFound):
            self.Profile('test')


class EmptyProfile(ProfileBase):
    """ Given an empty default profile. """
    def test_profile_with_no_args(self) -> None:
        """ Profile having no attrs should throw ProfileMissingArgs. """
        self.login_config = """
[default]
    """

        with self.assertRaises(ProfileMissingArgs):
            self.Profile()


class AttrTestMixin(ProfileBase):
    """ Mixin class of attribute tests valid for any profile. """

    def test_magic_attrs(self):
        """ Ensure AttributeError is thrown if a bad attr is referenced. """
        with self.assertRaises(AttributeError):
            getattr(self.profile, 'manbearbig_does_not_exit')

    def test_has_cookies_attr(self):
        """ Profile object should have a cookies attr. """
        self.assertTrue(
            hasattr(self.profile, 'cookies'),
            "Profile does not have the cookies attribute!"
        )

    def test_cookies_attr_in_dir(self):
        """ The cookies attribute should be visible in dir(profile). """
        self.assertIn(
            'cookies',
            dir(self.profile),
            "Cookies attribute not listed in dir(profile)!"
        )

    def test_has_pidfile_attr(self) -> None:
        """ Profile object should have a pidfile attr. """
        self.assertTrue(hasattr(self.profile, 'pidfile'))
        self.assertEqual(
            path.split(self.profile.pidfile)[1],
            self.profile.name + '.pid'
        )

    def test_raise_if_logged_in_no_pidfile(self) -> None:
        """ Should not throw exception if the pidfile does not exist. """
        if not hasattr(self.profile, 'pidfile'):
            raise SkipTest('Profile does not have attr pidfile!')
        if isfile(self.profile.pidfile):
            raise SkipTest('Test assumes pidfile is nonexistent!')

        self.profile.raise_if_logged_in()

    def test_raise_if_logged_in_with_good_pidfile(self) -> None:
        """ Should throw an exception if the pidfile exists. """
        if not hasattr(self.profile, 'pidfile'):
            raise SkipTest('Profile does not have attr pidfile!')
        with open(self.profile.pidfile, 'x') as f:
            f.write(str(getpid()))
        self.addCleanup(qremove, self.profile.pidfile)

        with self.assertRaises(AlreadyLoggedIn):
            self.profile.raise_if_logged_in()

    @patch('awscli_login.config.pid_exists', return_value=False)
    def test_raise_if_logged_in_with_bad_pidfile(self, mock) -> None:
        """ Should remove the pidfile if process does not exist. """
        pidfile = self.profile.pidfile
        with open(pidfile, 'x') as f:
            f.write(str(getpid()))
        self.addCleanup(qremove, pidfile)

        try:
            self.profile.raise_if_logged_in()
        finally:
            self.assertFalse(
                isfile(pidfile),
                "raise_if_logged_in failed to cleanup bad pidfile!"
            )


class ReadMinProfile(AttrTestMixin):
    """ Test profile with minimum valid default profile """

    def setUp(self) -> None:
        super().setUp()
        self.login_config = """
[default]
ecp_endpoint_url = foo
    """
        self.Profile()

    def test_simple_config(self) -> None:
        """ Simple config with just a SAML section. """
        self.assertProfileHasAttrs(ecp_endpoint_url='foo')

    def test_bad_attr(self):
        """ Ensure AttributeError is thrown if a bad attr is referenced. """
        profile = self.Profile()
        with self.assertRaises(AttributeError):
            getattr(profile, 'manbearbig_does_not_exit')

    def test_bad_magic_cookies(self):
        """ Ensure cookies returns None if username is not set! """
        profile = self.Profile()

        self.assertEqual(
            profile.cookies,
            None,
            'The username has not been set so cookies must be None!'
        )


class CookieMixin(TempDir):
    """
    Mixin class for testing awscli_login.config.Profile cookies.

    NOTA BENE: These tests assume that the username attribute has
    been set, and that self.profile has been set by setUp.
    """

    def _split_cookie_path(self):
        """ Returns a tuple (path, filename) for the cookie jar. """
        cookies = self.profile.cookies

        self.assertNotEqual(
            self.profile.username,
            None,
            'This test will fail if the username is not set!'
        )

        self.assertNotEqual(
            cookies,
            None,
            'The username is set so why no cookie?'
        )
        return path.split(cookies)

    def test_cookies_path(self):
        """ Cookie jar should live in ~/.aws-config/cookies. """
        self.Profile()

        directory, _ = self._split_cookie_path()
        jar_dir = path.join(self.tmpd.name, JAR_DIR)

        self.assertEqual(
            directory,
            jar_dir,
            "Expected cookies to be in directory %s" % JAR_DIR
        )

    def test_cookies_filename(self):
        """ Cookie jar format should be user.txt. """
        self.Profile()
        _, filename = self._split_cookie_path()

        self.assertEqual(
            filename,
            self.profile.username + '.txt',
            "Expected cookies filename to have format user.txt"
        )


class ReadFullProfile(ProfileBase, CookieMixin):
    """ Test default profile using all options """

    def setUp(self) -> None:
        super().setUp()
        self.login_config = """
[default]
ecp_endpoint_url = url
username = netid1
password = secret1
factor = push
role_arn = arn:aws:iam::account-id:role/role-name
enable_keyring = True
passcode = secret_code
verbose = 1
refresh = 1500
    """

    def test_full_config(self) -> None:
        """ Full config with User and SAML section """
        self.Profile()

        expected_attr_vals = {
            "ecp_endpoint_url": 'url',
            "username": 'netid1',
            "password": 'secret1',
            "factor": 'push',
            "role_arn": "arn:aws:iam::account-id:role/role-name",
            "enable_keyring": True,
            "passcode": "secret_code",
            "verbose": 1,
            "refresh": 1500,
        }

        self.assertProfileHasAttrs(**expected_attr_vals)


class ReadFullProfileTestOverrides(ReadFullProfile):
    """ Override profile with cli args """

    def test_full_config(self) -> None:
        """ Testing command line args are processed. """
        args = {
            "ecp_endpoint_url": 'url2',
            "username": 'netid2',
            "password": 'secret2',
            "factor": 'sms',
            "role_arn": "arn:aws:iam::account-id:role/role-name2",
            "passcode": "secret",
            "verbose": 2,
            "refresh": 1000,
        }  # Dict[str, Any]
        expected_attr_vals = copy(args)
        expected_attr_vals.update({'enable_keyring': False})
        args.update({'ask_password': True})

        self.Profile(profile='default', no_args=False, **args)
        self.assertProfileHasAttrs(**expected_attr_vals)


# This ensures that shared tests in mixins are not run with empty
# data sets!
del(CookieMixin)
del(AttrTestMixin)



import requests
import tempfile
import unittest

from datetime import datetime
from os import path
from os.path import dirname, abspath
from requests.exceptions import HTTPError
from unittest.mock import patch, MagicMock

from lxml.etree import XMLSyntaxError

from awscli_login.saml import (
    authenticate,
    authn_request,
    parse_role_arns,
    parse_soap_response,
    refresh,
    raise_if_saml_failed,
)

from awscli_login.exceptions import (
    AuthnFailed,
    InvalidSOAP,
    MissingCookieJar,
    RoleParseFail,
)
from awscli_login.util import (
    file2str,
    file2bytes,
)

DATA = path.join(dirname(abspath(__file__)), 'data')

ARN = 'arn:aws:iam::378517677616:'
SAML_SUCCESS = file2bytes(path.join(DATA, 'success.xml'))
SAML_SUCCESS_ARNS = [(
    ARN + 'saml-provider/shibboleth-test.techservices.illinois.edu',
    ARN + 'role/TestShibAdmin'
)]

SAML_SUCCESS_B64 = file2str(path.join(DATA, 'success.b64'))
SAML_AUTHNFAILED = file2bytes(path.join(DATA, 'error.xml'))
SAML_AUTHNFAILED_HTML = file2bytes(path.join(DATA, 'error.html'))
AUTHNREQUEST = file2bytes(path.join(DATA, 'authnrequest.xml'))


class saml(unittest.TestCase):
    """ Tests various SAML helper functions """

    def test_saml_authnfailed(self):
        """ raise_if_saml_failed should raise an error if AuthNfailed. """
        with self.assertRaises(AuthnFailed):
            raise_if_saml_failed(SAML_AUTHNFAILED)

    def test_saml_bad_soap(self):
        """ raise_if_saml_failed should raise an error if HTML is given. """
        with self.assertRaises(XMLSyntaxError):
            raise_if_saml_failed(SAML_AUTHNFAILED_HTML)

    def test_saml_success(self):
        """ raise_if_saml_failed should not raise an error on success. """
        raise_if_saml_failed(SAML_SUCCESS)

    @patch('awscli_login.saml.utcnow',
           return_value=datetime(2018, 4, 5, 16, 6, 50))
    @patch('awscli_login.saml.uuid4',
           return_value='37435E75-D889-45F3-A8EC-FD538C1F6BCC')
    def test_authn_request(self, *args):
        """ authn_request should return expected output with static inputs. """
        self.assertEqual(authn_request(), AUTHNREQUEST)

    def test_parse_soap_response(self):
        """ Given a SAML response PSR should return roles and assertion. """
        assertion, arns = parse_soap_response(SAML_SUCCESS)

        self.assertEqual(assertion, SAML_SUCCESS_B64)
        self.assertEqual(arns, SAML_SUCCESS_ARNS)

    def test_parse_role_arns(self):
        """ Given bad soap PRA should throw a SAML exception. """
        class role:
            text = 'bad role'

        with self.assertRaises(RoleParseFail):
            parse_role_arns([role(), role()])


class auth(unittest.TestCase):
    """ Tests for saml.authenticate and saml.refresh. """

    class MockRespone:
        """ A class for mocking a requests' response. """
        status_code = 0  # type: int
        content = None  # type: bytes

        def raise_for_status(self):
            pass

    class Success(MockRespone):
        status_code = 200
        content = SAML_SUCCESS

    class Failure401(MockRespone):
        status_code = 401
        content = SAML_AUTHNFAILED_HTML

        def raise_for_status(self):
            raise HTTPError()

    class FailureSoap(MockRespone):
        status_code = 200
        content = SAML_AUTHNFAILED

    class BadSoap(MockRespone):
        status_code = 200
        content = SAML_AUTHNFAILED_HTML

    def auth(self, cookies: str):
        """ Wrapper for calling saml.authenticate. """
        return authenticate(
            self.URL,
            cookies,  # TODO FIXME
            username='netid',
            password='password',
            headers={},
        )

    def refresh(self, cookies: str):
        """ Wrapper for calling saml.refresh. """
        return refresh(
            self.URL,
            cookies,
        )

    def setUp(self):
        idphost = "shibboleth-test.techservices.illinois.edu"

        self.URL = "https://" + idphost + "/idp/profile/SAML2/SOAP/ECP"
        self.cookies = 'cookies.txt'

    @patch('awscli_login.saml.Session')
    def auth_test(self, returns, saml, roles, test_func, mock,
                  create_cookies=False):
        """
        Helper function for testing saml authenticate & refresh.

        Tests that test_func returns the expected output.

        Args:
            returns (list): MockRespone(s) to be returned by requests.post.
            saml (str): Expected base 64 SAML assertion to be returned by
                        test_func.
            roles (Role): Expected roles to be returned by test_func.
            test_func: either self.auth or self.refresh.
        """
        mock.return_value = requests.Session()
        mock.return_value.post = MagicMock(side_effect=returns)

        with tempfile.TemporaryDirectory() as tempdir:
            cookies = path.join(tempdir, 'cookies.txt')
            if create_cookies:
                with open(cookies, 'w') as f:
                    f.write('#LWP-Cookies-2.0')

            encoded, arns = test_func(cookies)

        self.assertEqual(encoded, saml)
        self.assertEqual(arns, roles)

    def test_authenticate_with_username_password(self):
        """ Simulates successful auth with username/password. """
        idp = [self.Success()]
        self.auth_test(idp, SAML_SUCCESS_B64, SAML_SUCCESS_ARNS,
                       self.auth)

    def test_authenticate_bad_login_401(self):
        """ Simulates a failed user login. IdP returns 401 error. """
        idp = [self.Failure401()]

        with self.assertRaises(HTTPError):
            self.auth_test(idp, "", {}, self.auth)

    def test_authenticate_bad_login_soap(self):
        """ Simulates a failed user login. IdP returns soap error. """
        idp = [self.FailureSoap()]

        with self.assertRaises(AuthnFailed):
            self.auth_test(idp, "", {}, self.auth)

    def test_authenticate_bad_soap(self):
        """ Simulates a bad ECP endpoint that returns junk on auth. """
        idp = [self.BadSoap()]

        with self.assertRaises(InvalidSOAP):
            self.auth_test(idp, "", {}, self.auth)

    def test_refresh_with_cookies(self):
        """ Simulates successful refresh with cookies. """
        idp = [self.Success()]
        self.auth_test(idp, SAML_SUCCESS_B64, SAML_SUCCESS_ARNS, self.refresh,
                       create_cookies=True)

    def test_refresh_bad_login_401(self):
        """ Simulates failed refresh with cookies. IdP returns 401 error. """
        idp = [self.Failure401()]

        with self.assertRaises(HTTPError):
            self.auth_test(idp, "", {}, self.refresh, create_cookies=True)

    def test_refresh_bad_login_soap(self):
        """ Simulates failed refresh with cookies. IdP returns soap error. """
        idp = [self.FailureSoap()]

        with self.assertRaises(AuthnFailed):
            self.auth_test(idp, "", {}, self.refresh, create_cookies=True)

    def test_refresh_bad_soap(self):
        """ Simulates a bad ECP endpoint that returns junk on refresh. """
        idp = [self.BadSoap()]

        with self.assertRaises(InvalidSOAP):
            self.auth_test(idp, "", {}, self.refresh, create_cookies=True)

    def test_refresh_no_cookie_jar(self):
        """ Simulates a no cookie jar found error on refresh. """
        idp = [self.FailureSoap()]

        with self.assertRaises(MissingCookieJar):
            self.auth_test(idp, "", {}, self.refresh, create_cookies=False)


if __name__ == '__main__':
    suite = unittest.TestSuite()
    unittest.main()



""" Tests for profile.update the backend for aws configure login """
# from os.path import getmtime, relpath
# from time import sleep
from typing import Any
from unittest.mock import patch

from .base import ProfileBase
from .util import user_input


class UpdateProfileBase(ProfileBase):

    def assertProfileHasAttrsAfterUpdate(self, no_change=False,
                                         **kwargs: Any) -> None:
        """ Simulate user configuration of default profile. """
        usr_input = user_input(self.profile, kwargs)

        @patch('builtins.input', side_effect=usr_input)
        def profileUpdate(mock):
            self.profile.update()

        if no_change:
            self.assertTmpFileNotChangedBy(
                self.login_config_path,
                profileUpdate,
            )
        else:
            self.assertTmpFileChangedBy(
                self.login_config_path,
                profileUpdate,
            )

        self.profile.reload()
        try:
            self.assertProfileHasAttrs(**kwargs)
        except AssertionError as e:
            mesg = "\n\nUser input: " + ', '.join(usr_input) + '\n'
            mesg += "\nConfig file on disk: \n"
            with open(self.profile.config_file, 'r') as f:
                mesg += f.read()

            e.args = (e.args[0] + mesg, )
            raise e


class UpdateDefaultProfileTest(UpdateProfileBase):
    """ Test updating the default profile configuration """

    def setUp(self) -> None:
        super().setUp()
        self.login_config = """
[default]
ecp_endpoint_url = url
username = netid1
enable_keyring = True
role_arn = arn:aws:iam::account-id:role/role-name
factor = push
    """

    def test_profile_update(self) -> None:
        """Simulate user configuration of default profile. """
        self.Profile()
        self.assertProfileHasAttrsAfterUpdate(
            ecp_endpoint_url='url2',
            username='netid2',
            enable_keyring=False,
            role_arn="arn:aws:iam::account-id:role/role-name2",
            factor='auto',
        )

    def test_profile_update_no_change(self) -> None:
        """Simulate user making no changes with configuration tool. """
        self.Profile()
        self.assertProfileHasAttrsAfterUpdate(no_change=True)

    def test_non_default_profile_update(self) -> None:
        """Simulate user configuration of default profile. """
        self.login_config = self.login_config.replace(
            'default',
            'test'
        )

        self.Profile('test')
        self.assertProfileHasAttrsAfterUpdate(
            ecp_endpoint_url='url2',
            username='netid2',
            enable_keyring=False,
            role_arn="arn:aws:iam::account-id:role/role-name2",
            factor='auto',
        )

    def test_non_default_profile_update_no_change(self) -> None:
        """ Simulate user making no changes to non-default profile. """
        self.login_config = self.login_config.replace(
            'default',
            'test'
        )

        self.Profile('test')
        self.assertProfileHasAttrsAfterUpdate(no_change=True)



import os
import unittest

from datetime import datetime, timezone
from io import StringIO
from os.path import isfile
from typing import Any, Dict
from unittest.mock import patch

from botocore.session import Session

from awscli_login.const import ERROR_INVALID_PROFILE_ROLE
from awscli_login.exceptions import SAML
from awscli_login.util import (
    get_selection,
    remove_credentials,
    save_credentials,
    secure_touch,
    sort_roles,
)

from .base import CleanAWSEnvironment, TempDir


def token(akey: str, skey: str, stoken: str) -> Dict[str, Dict[str, Any]]:
    return {
               'Credentials': {
                                  'AccessKeyId': akey,
                                  'SecretAccessKey': skey,
                                  'SessionToken': stoken,
                                  'Expiration': datetime.now(timezone.utc)
                              }
           }


class util(unittest.TestCase):

    @patch('builtins.input', return_value=0)
    def test_get_single_selection(self, mock_input):
        """ When a single role is returned by the IdP do not ask for input """
        roles = [('idp', 'arn:aws:iam::224588347132:role/KalturaAdmin')]

        with patch('sys.stdout', new=StringIO()) as mock_stdout:
            self.assertEqual(get_selection(roles), roles[0])

            mock_input.assert_not_called()

            output = mock_stdout.getvalue()
            self.assertEqual(output, "", "get_selection printed output:"
                             "\n%s\n----\n" % output)

    @patch('builtins.input', return_value=0)
    @patch('sys.stdout', new=StringIO())
    def test_get_1of2_selections(self, *args):
        """ Select the first of two roles """
        roles = [
            ('idp1', 'arn:aws:iam::224588347132:role/KalturaAdmin'),
            ('idp2', 'arn:aws:iam::617683844790:role/BoxAdmin'),
        ]

        self.assertEqual(get_selection(roles), roles[0])

    @patch('builtins.input', return_value=1)
    @patch('sys.stdout', new=StringIO())
    def test_get_2of2_selections(self, *args):
        """ Select the second of two roles """
        roles = [
            ('idp1', 'arn:aws:iam::224588347132:role/KalturaAdmin'),
            ('idp2', 'arn:aws:iam::617683844790:role/BoxAdmin'),
        ]

        self.assertEqual(get_selection(roles), roles[1])

    @patch('builtins.input', return_value=1)
    def test_selections_profile_role(self, *args):
        """ Profile role is selected when valid and present """
        roles = [
            ('idp1', 'arn:aws:iam::224588347132:role/KalturaAdmin'),
            ('idp2', 'arn:aws:iam::617683844790:role/BoxAdmin'),
        ]
        profile_role = roles[1][1]

        with patch('sys.stdout', new=StringIO()) as stdout:
            role = get_selection(roles, profile_role)

            self.assertEqual(
                 stdout.getvalue(),
                 '',
                 'User was prompted for a selection '
                 'even though the Profile role was set!'
            )

        self.assertEqual(
            role,
            roles[1],
            'Profile role was not selected!'
        )

    @patch('builtins.input', return_value=1)
    def test_selections_bad_profile_role(self, *args):
        """ If a bad Profile role is set, then get_selection prompts the user.
        """
        profile_role = 'arn:aws:iam::617683844790:role/BadRole'

        roles = [
            ('idp1', 'arn:aws:iam::224588347132:role/KalturaAdmin'),
            ('idp2', 'arn:aws:iam::617683844790:role/BoxAdmin'),
        ]

        with patch('sys.stdout', new=StringIO()):
            with self.assertLogs('awscli_login.util', 'ERROR') as cm:
                get_selection(roles, profile_role)

        error = ERROR_INVALID_PROFILE_ROLE % profile_role
        self.assertEqual(
            cm.output,
            ["ERROR:awscli_login.util:%s" % error],
        )

    def test_get_empty_selection(self, *args):
        """ Attempt to select from an empty role set """
        with self.assertRaises(SAML):
            get_selection([])

    def test_sort_roles(self, *args):
        """ Sort role arns by account and role. """
        roles = [
            ('idp1', 'arn:aws:iam::617683844790:role/KalturaAdmin'),
            ('idp1', 'arn:aws:iam::224588347132:role/KalturaAdmin'),
            ('idp1', 'arn:aws:iam::224588347132:role/ASFoobarTeam'),
        ]

        expected = [
            ('224588347132', [(2, 'ASFoobarTeam'), (1, 'KalturaAdmin')]),
            ('617683844790', [(0, 'KalturaAdmin')]),
        ]

        output = sort_roles(roles)
        self.assertEqual(
            expected,
            output,
            'Accounts & roles were not sorted as expected!'
            '\nReturned: %s'
            '\nExpected: %s' % (output, expected)
        )


class SaveDefaultCreds(CleanAWSEnvironment):

    def test_save_credentials_default_profile(self):
        """ Creates a default entry in non-existent ~/.aws/credentials """
        self.profile = None  # default
        credentials = """[default]
aws_access_key_id = foo
aws_secret_access_key = bar
aws_session_token = yep
"""

        session = Session()
        save_credentials(session, token('foo', 'bar', 'yep'))
        self.assertAwsCredentialsEquals(credentials)

    def test_save_credentials_not_default_profile(self):
        """ Creates a non-default entry in empty ~/.aws/credentials """
        self.profile = 'wtf'
        credentials = """[wtf]
aws_access_key_id = a
aws_secret_access_key = b
aws_session_token = c
"""

        session = Session()
        save_credentials(session, token('a', 'b', 'c'))
        self.assertAwsCredentialsEquals(credentials)

    def test_remove_credentials_default_profile(self):
        """ Removes default entry in ~/.aws/credentials """
        self.profile = 'default'
        self.aws_credentials = """
[default]
aws_access_key_id = foo
aws_secret_access_key = bar
aws_session_token = yep
"""

        credentials = "\n[default]\n" \
                      "aws_access_key_id = \n" \
                      "aws_secret_access_key = \n" \
                      "aws_session_token = \n"

        session = Session()
        remove_credentials(session)
        self.assertAwsCredentialsEquals(credentials)


class TestDontClobberCommentsBase(CleanAWSEnvironment):

    def setUp(self):
        super().setUp()
        self.aws_credentials = """
# This is an example credentials files with comments
# and old entries

[default]
aws_access_key_id = abc
aws_secret_access_key = def
aws_session_token = ghi

# Test

[wtf]
aws_access_key_id = 123
aws_secret_access_key = 456
aws_session_token = 789

# Test
"""

    def test_save_credentials_comments_not_clobbered(self):
        """ Ensure comments are not clobbered in ~/.aws/credentials """
        self.profile = 'foo'

        credentials = self.aws_credentials
        credentials += """[foo]
aws_access_key_id = a
aws_secret_access_key = b
aws_session_token = c
"""
        session = Session()
        save_credentials(session, token('a', 'b', 'c'))
        self.assertAwsCredentialsEquals(credentials)

    def test_remove_credentials_non_default_profile(self):
        """ Removes non-default entry in ~/.aws/credentials """
        self.profile = 'foo'

        credentials = self.aws_credentials
        credentials += "\n[foo]\n" \
                       "aws_access_key_id = \n" \
                       "aws_secret_access_key = \n" \
                       "aws_session_token = \n"

        self.aws_credentials += """
[foo]
aws_access_key_id = a
aws_secret_access_key = b
aws_session_token = c
"""

        session = Session()
        remove_credentials(session)
        self.assertAwsCredentialsEquals(credentials)


class SecureTouchTests(TempDir):
    """ Tests for the function secure_touch. """

    def test_secure_touch_function_creates_file(self):
        """Newly created file by secure_touch should have perms 0x600. """
        path = self._abspath('foo.txt')
        secure_touch(path)

        self.assertTrue(isfile(path), 'Failed to create file!')
        self.assertHasFilePerms(path, owner='rw')

    def test_secure_touch_function_changes_perms(self):
        """File with perms 0x644 should be 0x600 after secure_touch. """
        path = self._abspath('foo.txt')

        fd = os.open(path, os.O_CREAT | os.O_RDONLY, mode=0o644)
        os.close(fd)

        secure_touch(path)
        self.assertHasFilePerms(path, owner='rw')


if __name__ == '__main__':
    suite = unittest.TestSuite()
    unittest.main()



""" Utility functions """
import contextlib

from os import remove
from typing import Any, Dict, List  # NOQA

from awscli_login.config import Profile


class MockSession:
    """
    Minimal mock session class needed for class Profile init.
    """
    profile = None  # type: str

    def __init__(self, profile: str = None) -> None:
        self.profile = profile


def qremove(filename: str):
    """
    Removes a file quietly i.e. no error thrown if the file does
    not exist.
    """
    with contextlib.suppress(FileNotFoundError):
        remove(filename)


def user_input(profile: Profile,
               expected_attr_vals: Dict[str, Any]) -> List[str]:
    """ Returns values that profile.update() modifies """
    r = []  # type: List[str]

    for key in profile._config_options.keys():
        r.append(str(expected_attr_vals.get(key, '')))
    return r



import os
import logging
import configparser

from datetime import datetime, timezone
from os import path
from time import sleep
from typing import Dict, List, Tuple

from awscli.customizations.configure.set import ConfigureSetCommand
from botocore.session import Session

from .const import ERROR_INVALID_PROFILE_ROLE
from .exceptions import SAML
from .typing import Role

## WHY CAN'T I IMPORT util or __main__ ??
awsconfigfile = path.join('.aws', 'credentials')
accountsfile = path.join(
                         os.path.expanduser("~"),
                         '.aws-login',
                         'accounts.ini'
                        )

identityfile = path.join(
                         os.path.expanduser("~"),
                         '.aws-login',
                         'identity.txt'
                        )
def read_account_names_file():
    account_names = {}
    if os.path.exists(accountsfile):
        config = configparser.ConfigParser()
        config.read(accountsfile)
        for key in config['ACCOUNT_NAMES']:
            account_names[key] = config['ACCOUNT_NAMES'][key]
    return account_names


def write_identity_file(account_id):
    account_names = read_account_names_file()
    with open(identityfile,'w') as f:
        print(account_names[account_id], file=f)

logger = logging.getLogger(__name__)

TRUE = ("yes", "true", "t", "1")
FALSE = ("no", "false", "f", "0")


def sort_roles(role_arns: List[Role]) \
               -> List[Tuple[str, List[Tuple[int, str]]]]:
    """ TODO """
    accounts = {}  # type: Dict[str, List[Tuple[int, str]]]
    r = []  # type: List[Tuple[str, List[Tuple[int, str]]]]

    for index, arn in enumerate(role_arns):
        acct = arn[1].split(':')[4]  # type: str
        role = arn[1].split(':')[5].split('/')[1]  # type: str

        role_list = accounts.get(acct, list())
        role_list.append((index, role))
        accounts[acct] = role_list

    for acct in sorted(accounts.keys()):
        accounts[acct].sort(key=lambda x: x[1])
        r.append((acct, accounts[acct]))

    return r


def get_selection(role_arns: List[Role], profile_role: str = None) -> Role:
    """ Interactively prompts the user for a role selection. """
    i = 0
    n = len(role_arns)
    select = {}  # type: Dict[int, int]

    # Return profile_role if valid and set
    if profile_role is not None:
        pr = [(idp, role) for idp, role in role_arns if role == profile_role]
        if pr:
            return pr[0]
        else:
            logger.error(ERROR_INVALID_PROFILE_ROLE % profile_role)

    if n > 1:
        print("Please choose the role you would like to assume:")
        account_names = read_account_names_file()

        accounts = sort_roles(role_arns)
        for acct, roles in accounts:
            if acct in account_names:
                print(
                        ' ' * 4,
                        "Account: %s (%s)" % (account_names[acct], acct)
                     )
            else:
                print(' ' * 4, "Account:", acct)

            for index, role in roles:
                print(' ' * 8, "[ %d ]:" % i, role)
                select[i] = index
                i += 1

        print("Selection:\a ", end='')
# TODO need error checking
        return role_arns[select[int(input())]]
    elif n == 1:
        return role_arns[0]
    else:
        raise SAML("No roles returned!")


class Args:
    """ A stub class for passing arguments to ConfigureSetCommand """
    def __init__(self, varname: str, value: str) -> None:
        self.varname = varname
        self.value = value


def _aws_set(session: Session, varname: str, value: str) -> None:
    """
    This is a helper function for save_credentials.

    The function is the same as running:

    $ aws configure set varname value
    """
    set_command = ConfigureSetCommand(session)
    set_command._run_main(Args(varname, value), parsed_globals=None)


def remove_credentials(session: Session) -> None:
    """
    Removes current profile's credentials from ~/.aws/credentials.
    """

    ConfigureSetCommand._WRITE_TO_CREDS_FILE.append("aws_security_token")
    profile = session.profile if session.profile else 'default'

    _aws_set(session, 'aws_access_key_id', '')
    _aws_set(session, 'aws_secret_access_key', '')
    _aws_set(session, 'aws_session_token',  '')
    _aws_set(session, 'aws_security_token', '')
    logger.info("Removed temporary STS credentials from profile: " + profile)

def remove_identity_file():
    os.remove(identityfile)

def save_credentials(session: Session, token: Dict) -> datetime:
    """ Takes an Amazon token and stores it in ~/.aws/credentials """

    ConfigureSetCommand._WRITE_TO_CREDS_FILE.append("aws_security_token")

    creds = token['Credentials']
    profile = session.profile if session.profile else 'default'

    _aws_set(session, 'aws_access_key_id', creds['AccessKeyId'])
    _aws_set(session, 'aws_secret_access_key', creds['SecretAccessKey'])
    _aws_set(session, 'aws_session_token',  creds['SessionToken'])
    _aws_set(session, 'aws_security_token',  creds['SessionToken'])
    logger.info("Saved temporary STS credentials to profile: " + profile)

    assert isinstance(creds['Expiration'], datetime), \
        "Amazon returned bad Expiration!"
    return creds['Expiration']


def file2bytes(filename: str) -> bytes:
    """
    Takes a filename and returns a byte string with the content of the file.
    """
    with open(filename, 'rb') as f:
        data = f.read()
    return data


def file2str(filename: str) -> str:
    """ Takes a filename and returns a string with the content of the file. """
    with open(filename, 'r') as f:
        data = f.read()
    return data


def nap(expires: datetime, percent: float) -> None:
    """TODO. """
    tz = timezone.utc
    ttl = int((expires - datetime.now(tz)).total_seconds())
    sleep_for = ttl * 0.9

    logger.info('Going to sleep for %d seconds.' % sleep_for)
    sleep(sleep_for)


def secure_touch(path):
    """Sets a file's permissions to read/write by owner only.

    Sets the file `path`'s mode to 600. The file `path` is created
    if it does not exist.

    Args:
        path - A path to a file.
    """
    fd = os.open(path, os.O_CREAT | os.O_RDONLY, mode=0o600)
    os.fchmod(fd, 0o600)
    os.close(fd)



from argparse import Namespace
from functools import wraps
from multiprocessing import Process, Pipe
from os import walk, sep
from os.path import basename, getmtime
from sys import exc_info
from time import sleep
from typing import Callable
from unittest.mock import patch

# This module allows a traceback to be pickled!
from tblib import pickling_support

from awscli.clidriver import main

from awscli_login.util import (
    file2bytes,
)


def fork():
    """A decorator for running a function in a separate forked process.

    This is particularly useful for running unittests in a separate
    process were a strong degree of isolation is needed. Exceptions
    from the subprocess are captured and reraised in the decorator
    so that unittests will function as expected.

    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            parent, child = Pipe()

            def error_handler(conn, func, *args, **kwargs):
                try:
                    func(*args, **kwargs)
                except Exception:
                    pickling_support.install()
                    etype, exp, tb = exc_info()
                    conn.send((etype, str(exp), tb))
                    exit(1)
                finally:
                    conn.close()

            p = Process(
                target=error_handler,
                args=(child, func, *args),
                kwargs=kwargs,
            )
            p.start()
            p.join()

            if p.exitcode == 1:
                etype, message, tb = parent.recv()
                raise Exception(str(etype) + ': ' + message).with_traceback(tb)
            return
        return wrapper
    return decorator


def login_cli_args(
    ask_password=False,
    ecp_endpoint_url=None,
    factor=None,
    force_refresh=False,
    passcode=None,
    password=None,
    refresh=0,
    role_arn=None,
    subcommand=None,  # TODO FIXME!?
    username=None,
    verbose=0,
) -> Namespace:
    """
    Function for mocking cli args to login.
    """
    return Namespace(**locals())


def configure_cli_args(
    verbose=0,
) -> Namespace:
    """
    Function for mocking cli args to login.
    """
    return Namespace(**locals())


def exec_awscli(*argv: str) -> None:
    """ Run the aws cli. """
    @patch('sys.argv', ['aws', *argv])
    def run_main():
        main()

    run_main()


# Based on code from:
# https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python
def tree(startpath):
    r = ''

    for root, dirs, files in walk(startpath):
        level = root.replace(startpath, '').count(sep)
        indent = ' ' * 4 * (level)
        r += '{}{}/\n'.format(indent, basename(root))
        subindent = ' ' * 4 * (level + 1)
        for f in files:
            r += '{}{}\n'.format(subindent, f)

    return r


def isFileChangedBy(filename: str, func: Callable,
                    *args, **kwargs) -> bool:
    """Tests if a file is changed by a function call.

    Args:
        filename: The file to test.
        func: The function to call
        *args: positional arguments to pass to `func`.
        **kwargs: keyword arguments to pass to `func`.

    Returns:
        True if the file was changed, False otherwise.
    """

    before = file2bytes(filename)
    func(*args, **kwargs)
    after = file2bytes(filename)

    return not (before == after)


def isFileTouchedBy(filename: str, func: Callable,
                    *args, **kwargs) -> bool:
    """Tests if a file is toched by a function call.

    Args:
        filename: The file to test.
        func: The function to call
        *args: positional arguments to pass to `func`.
        **kwargs: keyword arguments to pass to `func`.

    Returns:
        True if the file was touched, False otherwise.
    """
    sleep(1)  # Must sleep or before == after!

    before = getmtime(filename)
    func(*args, **kwargs)
    after = getmtime(filename)

    return not bool(after - before)



from .base import ProfileBase


class IsFactorPromptDisabled(ProfileBase):
    """
    Tests that profile.is_factor_prompt_disabled and
    profile.is_factor_valid work as expected
    """

    def setUp(self) -> None:
        super().setUp()
        self.login_config = """
[default]
ecp_endpoint_url = foo
    """
        self.Profile()

    # Profile.is_factor_prompt_disabled() tests

    def test_factor_none(self):
        "Factor prompt should be enabled when no factor is given. """
        # self.profile.factor = None
        self.assertFalse(self.profile.is_factor_prompt_disabled())

    def test_factor_off(self):
        "Factor prompt should be disabled when factor is disabled. """
        self.profile.factor = 'off'
        self.assertTrue(self.profile.is_factor_prompt_disabled())

        self.profile.factor = 'OFF'
        self.assertTrue(self.profile.is_factor_prompt_disabled(),
                        "Should be case insensitive")

    def test_factor_auto(self):
        "Factor prompt should be disabled when a valid factor is given. """
        self.profile.factor = 'auto'
        self.assertTrue(self.profile.is_factor_prompt_disabled())

    def test_factor_invalid(self):
        "Factor prompt should be enabled when an invalid factor is given. """
        self.profile.factor = 'bozo'
        self.assertFalse(self.profile.is_factor_prompt_disabled())

    # Profile.is_factor_valid() tests

    def test_is_factor_auto_valid(self):
        "Auto should be a valid factor. """
        self.profile.factor = 'auto'
        self.assertTrue(self.profile.is_factor_valid())

    def test_is_factor_push_valid(self):
        "Push should be a valid factor. """
        self.profile.factor = 'push'
        self.assertTrue(self.profile.is_factor_valid())

    def test_is_factor_bozo_valid(self):
        "Bozo should not be a valid factor. """
        self.profile.factor = 'bozo'
        self.assertFalse(self.profile.is_factor_valid())

    def test_is_factor_off_valid(self):
        "Off should not be a valid factor. """
        self.profile.factor = 'off'
        self.assertFalse(self.profile.is_factor_valid())

    def test_is_factor_None_valid(self):
        "None should not be a valid factor. """
        # self.profile.factor = None
        self.assertFalse(self.profile.is_factor_valid())



from botocore.session import Session
import boto3
from .__main__ import error_handler, save_sts_token
from .util import sort_roles
from .config import Profile
from .saml import (
    authenticate,
    refresh
)
import os
import configparser

accountsfile = os.path.join(
                         os.path.expanduser("~"),
                         '.aws-login',
                         'accounts.ini'
                        )


@error_handler()
def print_account_names(profile: Profile, session: Session):
    print(list_account_names(profile,session))


@error_handler()
def save_account_names(profile: Profile, session: Session):
    names = list_account_names(profile, session)
    write_account_names_file(names)


def write_account_names_file(account_names):
    config = configparser.ConfigParser()
    if os.path.exists(accountsfile):
        config.read(accountsfile)

    if not config.has_section("ACCOUNT_NAMES"):
        config.add_section("ACCOUNT_NAMES")

    for account_id in account_names:
        if not config.has_option("ACCOUNT_NAMES", account_id):
            config.set("ACCOUNT_NAMES",account_id, account_names[account_id])

    with open(accountsfile,'w') as f:
        config.write(f)

def list_account_names(profile: Profile, session: Session):
    """ Print account names to STDOUT """
    
    sts = boto3.client('sts')

    profile.get_username()

    try:
        saml, roles = refresh(
            profile.ecp_endpoint_url,
            profile.cookies,
        )
    except Exception:
        creds = profile.get_credentials()
        saml, roles = authenticate(profile.ecp_endpoint_url,
                                   profile.cookies, *creds)

    account_roles = [account_id for account_id, _ in sort_roles(roles)]
    discovered = {}

    for account_id, role  in list(zip(account_roles,roles)):
        save_sts_token(session, sts, saml, role, profile.duration)

        params = dict(
            RoleArn=role[1],
            PrincipalArn=role[0],
            SAMLAssertion=saml,
        )
        # duration is optional and can be set by the role;
        # avoid passing if not set.

        token = sts.assume_role_with_saml(**params)
        creds = token['Credentials']
        iam = boto3.client(
                'iam',
                aws_access_key_id=creds['AccessKeyId'],
                aws_secret_access_key=creds['SecretAccessKey'],
                aws_session_token=creds['SessionToken']
        )

        aliases = iam.list_account_aliases()
        default_alias = aliases['AccountAliases'][0]
        discovered[account_id] = default_alias

    return discovered



import os
import stat
import unittest

from os import environ, makedirs, path, walk, unlink
from os.path import dirname, relpath, join
from tempfile import TemporaryDirectory
from typing import Any, Callable, List  # NOQA
from unittest.mock import patch

from wurlitzer import pipes

from awscli_login.config import CONFIG_FILE

from .util import exec_awscli, isFileChangedBy, isFileTouchedBy, tree
from .exceptions import NotRelativePathError

PERMISSIONS = [
    ('r', 'read'),
    ('w', 'write'),
    ('x', 'execute')
]

FLAGS = {
    'Owner': {
        'r': stat.S_IRUSR,
        'w': stat.S_IWUSR,
        'x': stat.S_IXUSR,
    },
    'Group': {
        'r': stat.S_IRGRP,
        'w': stat.S_IWGRP,
        'x': stat.S_IXGRP,
    },
    'Others': {
        'r': stat.S_IRGRP,
        'w': stat.S_IWGRP,
        'x': stat.S_IWGRP,
    },
}


class SDGTestCase(unittest.TestCase):
    """Adds additional convenience assert statements.

    .. testsetup::

        from tests.base import SDGTestCase
        self = SDGTestCase
    """

    @staticmethod
    def assertHasFilePerms(path: str, owner: str = '', group: str = '',
                           others: str = '') -> None:
        info = os.stat(path)

        users = [
            ('Owner', owner),
            ('Group', group),
            ('Others', others),
        ]

        for user, perms in users:
            for p, perm in PERMISSIONS:
                hasPerm = bool(info.st_mode & FLAGS[user][p])

                if p in perms and not hasPerm:
                    raise AssertionError(
                        '%s does not have %s permission!' % (user, perm)
                    )
                elif p not in perms and hasPerm:
                    raise AssertionError(
                        '%s has %s permission!' % (user, perm)
                    )

    @staticmethod
    def _assertHasAttr(obj: object, attr: str, evalue: Any) -> str:
        """Assert object has an attribute with a given value.

        Example:
            This is an example.

        Args:
            obj: object to test.
            attr: attribute name.
            evalue: expected value.

        Returns:
            Error message or None.
        """
        objName = obj.__class__.__name__

        try:
            rvalue = getattr(obj, attr)
        except AttributeError:
            return "%s object does not have attr: %s" % (objName, attr)

        if type(rvalue) != type(evalue):
            return "Attribute %s has unexpected type: %s\n" \
                   "Expected %s on %s object!" % \
                   (attr, type(rvalue), type(evalue), object)

        if rvalue != evalue:
            return "Attribute %s has unexpected value: %s\n" \
                   "Expected %s on %s object!" % \
                   (attr, rvalue, evalue, objName)

        return None

    @staticmethod
    def assertHasAttr(obj: object, attr: str, evalue: Any) -> None:
        """Assert object has an attribute with a given value.

        Args:
            obj: object to test.
            attr: attribute name.
            evalue: expected value.

        Raises:
            AssertionError

        Examples:
            Consider the following simple class:

            >>> class Coordinates:
            ...     x = 0
            ...     y = 0
            ...
            >>> center = Coordinates()

            A simple passing test is the following:

            >>> self.assertHasAttr(center, 'x', 0)

            The following are various failing tests and the error's
            they raise:

            >>> self.assertHasAttr(center, 'foo', 'bar')
            Traceback (most recent call last):
                ...
            AssertionError: Coordinates object does not have attr: foo

            >>> self.assertHasAttr(center, 'x', 'bar')
            Traceback (most recent call last):
                ...
            AssertionError: Attribute x has unexpected type <class 'int'>
            Expected <class 'str'> on Coordinates object!

            >>> self.assertHasAttr(center, 'x', 1)
            Traceback (most recent call last):
                ...
            AssertionError: Attribute x has unexpected value 0
            Expected 1 on Coordinates object!
        """
        error = SDGTestCase._assertHasAttr(obj, attr, evalue)

        if error:
            raise AssertionError(error)

    @staticmethod
    def assertHasAttrs(obj: object, **kwargs: Any) -> None:
        """Assert object has attributes with given values.

        Tests if object `obj` has all the given named arguments as
        attributes with the corresponding named argument values.

        Args:
            obj: object to test.
            **kwargs: each keyword argument's name is expected to
                be an attribute on object `obj` having the same value
                as the keyword argument.

        Raises:
            AssertionError

        Examples:
            Consider the following simple class:

            >>> class Coordinates:
            ...     x = 0
            ...     y = 0
            ...
            >>> center = Coordinates()

            A simple passing test is the following:

            >>> self.assertHasAttrs(center, x=0, y=0)

            The following are various failing tests and the error's
            they raise:

            >>> self.assertHasAttrs(center, foo='bar', x=0, y=0)
            Traceback (most recent call last):
                ...
            AssertionError: Coordinates object does not have attr: foo

            >>> self.assertHasAttrs(center, x='bar', y='foo')
            Traceback (most recent call last):
                ...
            AssertionError: Attribute x has unexpected type <class 'int'>
            Expected <class 'str'> on Coordinates object!

            >>> self.assertHasAttrs(center, x=1, y=0)
            Traceback (most recent call last):
                ...
            AssertionError: Attribute x has unexpected value 0
            Expected 1 on Coordinates object!
        """
        errors = []  # type: List[str]

        for attr, value in kwargs.items():
            error = SDGTestCase._assertHasAttr(obj, attr, value)

            if error:
                errors.append(error)

        if len(errors):
            raise AssertionError('\n'.join(errors))


class TempDir(SDGTestCase):
    """A class for tests that require a temporary directory.

    This class creates a temporary directory during the setUp phase
    before each test is run. The temporary directory is destroyed
    after the test finishes even if the test failes.

    Attributes:
        tmpd (:obj:`tempfile.TemporaryDirectory`): The `TemporaryDirectory`
            object created.
    """

    def setUp(self) -> None:
        """Creates an empty temporary directory. """
        self.tmpd = TemporaryDirectory()
        self.addCleanup(self.tmpd.cleanup)

    def _abspath(self, rel_file_path) -> str:
        """Returns an absolute path rooted in a temporary directory.

        Args:
            rel_file_path: A relative file path.

        Returns:
            The absolute path to `tmpd/rel_file_path`.

        Raises:
            NotRelativePathError: If `rel_file_path` is an absolute path.
        """
        if path.isabs(rel_file_path):
            raise NotRelativePathError(rel_file_path)
        return join(self.tmpd.name, rel_file_path)

    def read(self, rel_file_path: str) -> str:
        """Reads a temporary file.

        Args:
            rel_file_path: A relative file path.

        Returns:
            The contents of file `rel_file_path` rooted in temporary
            directory `tmpd.name`.

        Raises:
            NotRelativePathError: If `rel_file_path` is an absolute path.
        """
        filename = self._abspath(rel_file_path)

        try:
            with open(filename, 'r') as f:
                value = f.read()

            return value
        except FileNotFoundError:
            return None

    def write(self, rel_file_path: str, value) -> None:
        """Writes a given value to a temporary file.

        The file is rooted in a temporary directory using the given
        relative path `rel_file_path`. If `value` is None the file
        will be removed from disk, otherwise the value is written
        to disk.

        Args:
            rel_file_path: A relative file path.
            value: String to write to disk.

        Raises:
            NotRelativePathError: If `rel_file_path` is an absolute path.
        """
        filename = self._abspath(rel_file_path)

        if value is not None:
            with open(filename, 'w') as f:
                f.write(value)
        else:
            try:
                unlink(filename)
            except FileNotFoundError:
                pass

    def dump(self):
        """Dumps temporary directory contents into an exception.

        Raises an exception containing a listing of the temporary directory
        contents, followed by the file contents, and the current enviroment.
        Useful for debugging.

        Raises:
            AssertionError
        """
        errors = '\n---------- Temporary Directory Listing ---------\n'
        errors += tree(self.tmpd.name)
        errors += '\n------------ File Contents --------------------\n'

        for (dirpath, dirnames, filenames) in walk(self.tmpd.name):
            path = relpath(dirpath, self.tmpd.name)

            for fn in filenames:
                errors += join(path, fn) + ':\n'
                with open(join(dirpath, fn), 'r') as f:
                    errors += f.read() + '\n---\n'

        errors += '-------------- Enviroment -----------------------\n'

        errors += '--- Enviroment ---\n'
        for key, value in environ.items():
            errors += "%s=%s\n" % (key, value)

        errors += '\n-----------------------------------------------\n'
        raise AssertionError(errors)

    def assertTmpFileChangedBy(self, filename: str, func: Callable,
                               *args, **kwargs) -> None:
        """Asserts file is changed by a function.

        Args:
            filename: The file to test.
            func: The function to call.
            *args: positional arguments to pass to `func`.
            **kwargs: keyword arguments to pass to `func`.

        Raises:
            AssertionError if file is not changed.
        """
        if not isFileChangedBy(filename, func, *args, **kwargs):
            raise AssertionError('File was not changed: %s' %
                                 relpath(filename, self.tmpd.name))

    def assertTmpFileNotChangedBy(self, filename: str, func: Callable,
                                  *args, **kwargs) -> None:
        """Asserts file is not changed by a function.

        Args:
            filename: The file to test.
            func: The function to call.
            *args: positional arguments to pass to `func`.
            **kwargs: keyword arguments to pass to `func`.

        Raises:
            AssertionError if file is changed.
        """
        if isFileChangedBy(filename, func, *args, **kwargs):
            raise AssertionError('File was changed: %s' %
                                 relpath(filename, self.tmpd.name))

    def assertTmpFileTouchedBy(self, filename: str, func: Callable,
                               *args, **kwargs) -> None:
        """Asserts file is touched by a function.

        Args:
            filename: The file to test.
            func: The function to call.
            *args: positional arguments to pass to `func`.
            **kwargs: keyword arguments to pass to `func`.

        Raises:
            AssertionError if file is not touched.
        """
        if not isFileTouchedBy(filename, func, *args, **kwargs):
            raise AssertionError('File was not touched: %s' %
                                 relpath(filename, self.tmpd.name))

    def assertTmpFileNotTouchedBy(self, filename: str, func: Callable,
                                  *args, **kwargs) -> None:
        """Asserts file is not touched by a function.

        Args:
            filename: The file to test.
            func: The function to call.
            *args: positional arguments to pass to `func`.
            **kwargs: keyword arguments to pass to `func`.

        Raises:
            AssertionError if file is touched.
        """
        if isFileTouchedBy(filename, func, *args, **kwargs):
            raise AssertionError('File was touched: %s' %
                                 relpath(filename, self.tmpd.name))


class CleanAWSLoginEnvironment(TempDir):
    """Sets up a clean environment for testing aws-login Profiles.

    This works by creating a temporary directory and setting it up
    as the home directory to be used by `awscli_login.config.Profile`.
    """

    @property
    def login_config(self) -> str:
        """Contents of `tmpd/.aws-login/config`. """
        return self.read(CONFIG_FILE)

    @login_config.setter
    def login_config(self, value: str) -> None:
        self.write(CONFIG_FILE, value)

    @property
    def login_config_path(self) -> str:
        """Full path to `tmpd/.aws-login/config`. """
        return self._abspath(CONFIG_FILE)

    def setUp(self) -> None:
        """Creates `tmpd/.aws-login/config` and patches `awscli_login.config` to
        use it. """
        super().setUp()
        makedirs(dirname(self.login_config_path), 0o700)

        patcher = patch(
                'awscli_login.config.expanduser',
                return_value=self.tmpd.name
        )
        self.home = patcher.start()
        self.addCleanup(patcher.stop)


class CleanAWSEnvironment(TempDir):
    """Class for testing the AWS CLI.

    Sets up a clean test environment for working with the AWS CLI
    in a temporary directory.
    """

    @property
    def profile(self) -> str:
        """Contents of environment variable AWS_PROFILE. """
        return environ.get('AWS_PROFILE')

    @profile.setter
    def profile(self, value: str) -> None:
        self._set_environ('AWS_PROFILE', value)

    AWS_CONFIG_FILE = '.aws/config'

    @property
    def aws_config(self) -> str:
        """Contents of the AWS config file. """
        return self.read(CleanAWSEnvironment.AWS_CONFIG_FILE)

    @aws_config.setter
    def aws_config(self, value: str) -> None:
        self.write(CleanAWSEnvironment.AWS_CONFIG_FILE, value)

    @property
    def aws_config_path(self) -> str:
        """Full path to AWS config file. """
        return self._abspath(self.AWS_CONFIG_FILE)

    AWS_CREDENTIALS_FILE = '.aws/credentials'

    @property
    def aws_credentials(self) -> str:
        """Contents of the AWS shared credentials file. """
        return self.read(CleanAWSEnvironment.AWS_CREDENTIALS_FILE)

    @aws_credentials.setter
    def aws_credentials(self, value: str) -> None:
        self.write(CleanAWSEnvironment.AWS_CREDENTIALS_FILE, value)

    @property
    def aws_credentials_path(self) -> str:
        """Full path to the AWS shared credentials file. """
        return self._abspath(self.AWS_CREDENTIALS_FILE)

    def assertAwsCredentialsEquals(self, credentials: str) -> None:
        """Assert AWS shared credentials file equals given value.

        Args:
            credentials: Value to compare AWS shared credentials file to.

        Raises:
            AssertionError
        """
        aws_credentials = self.aws_credentials
        self.assertEqual(
            aws_credentials,
            credentials,
            'Did not expect:\n' + aws_credentials + '---',
        )

    def _set_environ(self, key: str, value) -> None:
        """Set key in environment to value.

        Note:
            If value is None the key is removed from the
            environment.
        """
        if value is not None:
            environ[key] = value
        else:
            environ.pop(key, None)

    def _clear_environ(self, name: str) -> None:
        """Removes an environment variable from the environment.

        Ensures that after the test has finished the environment
        is restored, even if the test failes with an exception.
        """
        try:
            value = environ.pop(name)
        except KeyError:
            self.addCleanup(environ.pop, name, None)
        else:
            self.addCleanup(environ.update, {name: value})

    def setUp(self) -> None:
        """Creates an empty AWS environment in a temporary directory.

        Clears all AWS environment variables, and sets AWS_CONFIG_FILE
        and AWS_SHARED_CREDENTIALS_FILE to the empty temporary
        directory `tmpd/.aws`.

        Note:
            All cleared environment variables are restored after tearDown.
        """
        super().setUp()
        makedirs(dirname(self.aws_config_path), 0o700)

        # Clear all Amazon environment variables
        # https://docs.aws.amazon.com/cli/latest/topic/config-vars.html

        # General Options
        self._clear_environ('AWS_CONFIG_FILE')
        self._clear_environ('AWS_PROFILE')
        self._clear_environ('AWS_DEFAULT_REGION')
        self._clear_environ('AWS_DEFAULT_OUTPUT')
        self._clear_environ('AWS_CA_BUNDLE')

        # Credentials
        self._clear_environ('AWS_SHARED_CREDENTIALS_FILE')
        self._clear_environ('AWS_ACCESS_KEY_ID')
        self._clear_environ('AWS_SECRET_ACCESS_KEY')
        self._clear_environ('AWS_SESSION_TOKEN')
        self._clear_environ('AWS_METADATA_SERVICE_TIMEOUT')
        self._clear_environ('AWS_METADATA_SERVICE_NUM_ATTEMPTS')

        # Tell the awscli to use tmpd instead of $HOME
        self._set_environ('AWS_CONFIG_FILE', self.aws_config_path)
        self._set_environ('AWS_SHARED_CREDENTIALS_FILE',
                          self.aws_credentials_path)


class CleanTestEnvironment(CleanAWSLoginEnvironment, CleanAWSEnvironment):
    """Sets up a clean test environment in a temporary directory. """


class IntegrationTests(CleanTestEnvironment):

    def setUp(self):
        """Configures plugin to work with the awscli. """
        super().setUp()

        # TODO Should this be split out? TODO THIS SHOULD BE A TEST!
        # exec_awscli('configure', 'set', 'plugins.login', 'awscli_login')
        self.aws_config = '[plugins]\nlogin = awscli_login\n'

    def assertAwsCliReturns(self, *args, stdout='', stderr='', code=0):
        """Runs awscli and tests the output.

        This test runs the awscli command with `*args`. It tests
        that the exit code and stdout and stderr match the given
        expect values.

        Example:
            `self.assertAwsCliReturns('logout')` is equivalent to running:

                $ aws logout

            The test in this example will pass if `aws logout` returns
            exit code 0 and emits no output to stdout or stderr.
            Otherwise an AssertionError is raised.

        Args:
            *args (str): Arguments to be passed to the AWS command
                line utility.
            stdout (str): Expected output to stdout.
            stderr (str): Expected output to stderr.
            code (int): Expected return code.

        Raises:
            AssertionError
        """
        try:
            with pipes() as (out, err):
                with self.assertRaises(SystemExit) as e:
                    exec_awscli(*args)

            import sys
            cmd = ' '.join(sys.argv)
            mesg = "Error: ran '%s', on %s expected output: %s"

            test_out = out.read()
            test_err = err.read()

            self.assertEqual(
                test_out,
                stdout,
                mesg % (cmd, 'stdout', stdout)
            )
            self.assertEqual(
                test_err,
                stderr,
                mesg % (cmd, 'stderr', stderr)
            )
            self.assertEqual(
                e.exception.code,
                code,
                "Error: '%s' returned %d, expected: %d" %
                (cmd, e.exception.code, code)
            )

            # Must be closed explicitly otherwise we receive warnings
        finally:
            out.close()
            err.close()



from unittest.mock import call, patch

from botocore.session import Session

from awscli_login.configure import configure
from .base import CleanTestEnvironment
from .util import login_cli_args


# TODO Not very dry...
class TestNoProfile(CleanTestEnvironment):
    """ Integration tests for no profile. """
    profile = None  # default

    @patch('builtins.input', return_value='')
    def test_save_credentials_default_profile(self, mock_input):
        """ Creates a default entry in ~/.aws/credentials """
        args = login_cli_args()
        session = Session()
        with self.assertRaises(SystemExit) as e:
            configure(args, session)

        self.assertEqual(e.exception.code, 0)
        mock_input.assert_has_calls(
            [
                call('ECP Endpoint URL [None]: '),
                call('Username [None]: '),
                call('Enable Keyring [False]: '),
                call('Duo Factor [None]: '),
                call('Role ARN [None]: '),
            ],
        )
        self.assertEqual(len(mock_input.call_args_list), 5)



""" This module is used to process ~/.aws-login/config """
import logging

from argparse import Namespace
from collections import OrderedDict
from configparser import ConfigParser, SectionProxy
from getpass import getuser, getpass
from os import path, makedirs, unlink
from os.path import expanduser, isfile
from typing import Any, Dict, FrozenSet  # NOQA
from urllib.parse import urlparse

from awscli.customizations.configure.writer import ConfigFileWriter
from botocore.session import Session
from keyring import get_password, set_password
from psutil import pid_exists

from .exceptions import (
    AlreadyLoggedIn,
    InvalidFactor,
    ProfileMissingArgs,
    ProfileNotFound,
)
from .typing import Creds
from .util import secure_touch

CONFIG_DIR = '.aws-login'
CONFIG_FILE = path.join(CONFIG_DIR, 'config')
JAR_DIR = path.join(CONFIG_DIR, 'cookies')
LOG_DIR = path.join(CONFIG_DIR, 'log')

ERROR_NONE = 0
ERROR_UNKNOWN = 1

FACTORS = ['auto', 'push', 'passcode', 'sms', 'phone']
DISABLE = ["0", "no", "false", "off", "disable"]

logger = logging.getLogger(__name__)


class Profile:
    """
    This class reads the current login profile from ~/.aws-login/config
    """
    # Public vars
    name = None  # type: str  # Profile name

    # Required args from profile
    ecp_endpoint_url = None  # type: str

    # Optional args from profile
    username = None  # type: str
    password = None  # type: str
    role_arn = None  # type: str
    enable_keyring = False  # type: bool
    factor = None  # type: str
    passcode = None  # type: str
    verbose = 0  # type: int
    refresh = 3000  # type: int
    force_refresh = False  # type: bool
    duration = 0  # type: int
    disable_refresh = False  # type: bool

    # path to profile configuration file
    config_file = None  # type: str

    # Private vars
    _args = None  # type: Namespace
    _required = frozenset(['ecp_endpoint_url'])  # type: FrozenSet[str]
    _optional = {
            'username': None,
            'password': None,
            'role_arn': None,
            'enable_keyring': False,
            'factor': None,
            'passcode': None,
            'verbose': 0,
            'refresh': 3000,  # in seconds (every 50 mins)
            'force_refresh': False,
            'duration': 0,  # duration can't be less than 900, btw
            'disable_refresh': False
    }  # type: Dict[str, Any]

    _config_options = OrderedDict(
        [
            ('ecp_endpoint_url', 'ECP Endpoint URL'),
            ('username', 'Username'),
            ('enable_keyring', 'Enable Keyring'),
            ('factor', 'Duo Factor'),
            ('role_arn', 'Role ARN'),
        ]
    )  # type: Dict[str, str]

    # Extra override args from command line
    _override = {
        'enable_keyring': 'ask_password',
    }  # type: Dict[str, str]

    def _init_dir(self) -> None:
        """ Create ~/.aws-login directory if it does not exist. """
        home = expanduser('~')

        self.home = home
        self.config_file = path.join(self.home, CONFIG_FILE)

        makedirs(path.join(home, CONFIG_DIR), mode=0o700, exist_ok=True)
        makedirs(path.join(home, LOG_DIR), mode=0o700, exist_ok=True)
        makedirs(path.join(home, JAR_DIR), mode=0o700, exist_ok=True)

        self.pidfile = path.join(home, CONFIG_DIR, self.name + '.pid')
        self.logfile = path.join(home, LOG_DIR, self.name + '.log')

    def _set_attrs(self, validate: bool) -> None:
        """ Load login profile from configuration. """
        config = ConfigParser()
        config.read(self.config_file)
        logger.info("Loaded config file: " + self.config_file)

        # Requried configuration
        self._set_req_attrs(config, validate)

        # Optional configuration
        self._set_opt_attrs(config, validate)

        # Warn if unknown attributes are found
        self._warn_on_unknown_attrs(config, validate)

    def _set_attrs_from_args(self) -> None:
        """ Load command line options. """
        self.options = self._required | frozenset(self._optional.keys())
        options = self.options - frozenset(self._override.keys())

        for option in options:
            value = getattr(self._args, option)

            if value:
                setattr(self, option, value)

    def _set_override_attrs(self) -> None:
        """ Disable attrs if overriden on command line. """
        for attr, option in self._override.items():
            value = getattr(self._args, option)

            if value:
                setattr(self, attr, False)

    def __init__(self, session: Session, args: Namespace,
                 validate: bool = True) -> None:
        """Load login profile.

        Args:
            session: a botocore session used to determine the current profile.
            args: an object containing command line arguments.
            validate: if true will validate the profile.

        Raises:
            ProfileNotFound: If the Profile can not be found.
            ProfileMissingArgs: If required arguments are missing.
        """
        self.name = session.profile if session.profile else 'default'
        self._args = args

        self._init_dir()
        self.reload(validate)

        logger.info("Loaded login profile: " + self.name)

    def __getattr__(self, item):
        """ Process dynamic attributes. """
        if item == 'cookies':
            if self.username:
                filename = self.username + '.txt'
                return path.join(self.home, JAR_DIR, filename)
            else:
                return None
        else:
            mesg = "'%s' object has no attribute '%s'"
            raise AttributeError(mesg % (self.__class__.__name__, item))

    def __dir__(self):
        """ Allows dir to work with dynamic attributes. """
        return super().__dir__() + ['cookies']

    def raise_if_logged_in(self) -> None:
        """ Throws an exception if already logged in. """
        if isfile(self.pidfile):
            with open(self.pidfile, 'r') as f:
                pid = int(f.read())

            if pid_exists(pid):
                raise AlreadyLoggedIn
            else:
                unlink(self.pidfile)

    def _get_profile(self, config: ConfigParser,
                     validate: bool) -> SectionProxy:
        """ Helper function for grabing a profile. """
        if self.name not in config:
            if validate:
                raise ProfileNotFound(self.name)

            return None

        return config[self.name]

    def _set_req_attrs(self, config: ConfigParser, validate: bool) -> None:
        """ Load required args from profile [~/.aws-login/config]. """
        errors = []
        section = self._get_profile(config, validate)

        for attr in self._required:
            try:
                if validate:
                    setattr(self, attr, section[attr])
                elif section is not None:
                    setattr(self, attr, section.get(attr))
                else:
                    setattr(self, attr, None)
            except KeyError:
                errors.append(attr)
        if errors:
            raise ProfileMissingArgs(self.name, *errors)

    def _set_opt_attrs(self, config: ConfigParser, validate: bool) -> None:
        """ Load optional args from profile [~/.aws-login/config]. """
        value = None  # type: Any
        section = self._get_profile(config, validate)

        for attr, default in self._optional.items():
            # section.get(attr) will always be a string
            # because ConfigParser treats everything as
            # a string
            if section is not None:
                value = section.get(attr, default)
            else:
                value = default

            # Type cast string to correct type
            # based on the default value
            if value != default:
                if type(default) == bool:
                    value = section.getboolean(attr)
                elif type(default) == int:
                    value = int(value)

            setattr(self, attr, value)

    def _warn_on_unknown_attrs(self, config: ConfigParser,
                               validate: bool) -> None:
        """ Load required args from profile [~/.aws-login/config]. """
        section = self._get_profile(config, validate)

        for attr in section:
            if attr not in self._required and attr not in self._optional:
                logger.warn('Unknown attribute "' + attr + '" in ' +
                            self.name + ' profile ' + self.name)

    def is_factor_valid(self):
        """ Return True if self.factor is valid. False otherwise. """
        return self.factor in FACTORS

    def raise_if_factor_invalid(self):
        """ Raise exception if self.factor is invalid. """
        if self.factor and not self.is_factor_valid():
            raise InvalidFactor(self.factor)

    def is_factor_prompt_disabled(self) -> bool:
        return bool(self.factor) and (self.factor.lower() in DISABLE or
                                      self.factor in FACTORS)

    def get_username(self):
        """ Get username from user if necessary. """
        if self.username is None:
            username = getuser()
            self.username = input("Username [%s]: " % username) or username

    # Can we make this DRYer?
    def get_credentials(self, first_pass: bool = True) -> Creds:
        """ Get credentials from user if necessary. """
        self.get_username()

        password = None
        if self.enable_keyring:
            ukey = self.username + '@' + urlparse(self.ecp_endpoint_url).netloc
            password = get_password("awscli_login", ukey)
        if password is None:
            password = getpass()
        if self.enable_keyring:
            set_password("awscli_login", ukey, password)
        self.password = password

        # TODO add support for any factor...
        # https://github.com/JohnPfeifer/duo-non-browser/wiki
        # https://duo.com/docs/authapi#/auth
        headers = {}
        if not self.is_factor_prompt_disabled():
            self.factor = input('Factor: ') or None
            self.raise_if_factor_invalid()

        if self.is_factor_valid():
            headers['X-Shiboleth-Duo-Factor'] = self.factor
            headers['X-Shibboleth-Duo-Factor'] = self.factor
            if not first_pass or self.factor == 'passcode':
                headers['X-Shiboleth-Duo-Passcode'] = input('Code: ')
                headers['X-Shibboleth-Duo-Passcode'] = \
                    headers['X-Shiboleth-Duo-Passcode']

        return self.username, self.password, headers

    # TODO Add validation...
    def update(self) -> None:
        """ Interactively update the profile. """
        new_values = {}
        writer = ConfigFileWriter()

        for attr, string in self._config_options.items():
            value = getattr(self, attr, self._optional.get(attr))

            prompt = "%s [%s]: " % (string, value)
            value = input(prompt)

            if value:
                new_values[attr] = value

        if new_values:
            if self.name != 'default':
                new_values['__section__'] = self.name

            secure_touch(self.config_file)
            writer.update_config(new_values, self.config_file)

    def reload(self, validate: bool = True):
        """ Reloads profile from disk [~/.aws-login/config]. """
        self._set_attrs(validate)

        if self._args:
            self._set_attrs_from_args()
            self._set_override_attrs()



import logging

from logging import Logger
from typing import Tuple


def _cli_options(verbosity: int) -> Tuple[int, str]:
    fmt = '%(message)s'
    level = logging.WARN

    if verbosity == 1:
        level = logging.INFO

        urllib3 = logging.getLogger("urllib3")
        urllib3.setLevel(logging.DEBUG)

    if verbosity == 2:
        level = logging.DEBUG

        botocore = logging.getLogger("botocore")
        botocore.setLevel(logging.INFO)

        botocore = logging.getLogger("botocore.endpoint")
        botocore.setLevel(logging.DEBUG)

    if verbosity == 3:
        fmt = '%(name)s %(message)s'
        level = logging.NOTSET

    if verbosity > 3:
        fmt = '%(filename)s:%(lineno)d:%(name)s:%(message)s'
        level = logging.NOTSET

    return level, fmt


def configConsoleLogger(verbosity: int) -> None:
    level, fmt = _cli_options(verbosity)
    logging.basicConfig(format=fmt)

    root = logging.getLogger()
    root.setLevel(level)

    _commonLogger()
    return


def configFileLogger(fname: str, level: int) -> Logger:
    _commonLogger()

    logger = logging.getLogger()
    logger.setLevel(level)

    logFormatter = logging.Formatter(
        '%(asctime)s:%(levelname)s:%(message)s'
    )
    fileHandler = logging.FileHandler(fname)
    fileHandler.setFormatter(logFormatter)
    # fileHandler.setLevel(level)
    logger.addHandler(fileHandler)

    return logger


def _commonLogger() -> None:
    requests_log = logging.getLogger("urllib3")
    requests_log.propagate = True



# Rudimentary documentation for the aws-cli plugin API can be found
# here: https://github.com/aws/aws-cli/issues/1261
import logging

from argparse import Namespace

from awscli.customizations.commands import BasicCommand
from botocore.session import Session

from .__main__ import main, logout
from .configure import configure
from .account_names import print_account_names, save_account_names

logger = logging.getLogger(__package__)


def awscli_initialize(cli):
    """ Entry point called by awscli """
    cli.register('building-command-table.main', inject_commands)
    cli.register('building-command-table.login', inject_subcommands)


def inject_commands(command_table, session: Session, **kwargs):
    """
    Used to inject top-level commands in the awscli command list.
    """
    command_table['login'] = Login(session)
    command_table['logout'] = Logout(session)


def inject_subcommands(command_table, session: Session, **kwargs):
    """
    Used to inject subcommands into the aws login command list.
    """
    command_table['configure'] = Configure(session)
    command_table['account_names'] = AccountNames(session)


class Login(BasicCommand):
    NAME = 'login'
    DESCRIPTION = ('is a plugin that manages retrieving and rotating'
                   ' Amazon STS keys using the Shibboleth IdP and Duo'
                   ' for authentication.')
    SYNOPSIS = ('aws login [<Arg> ...]')

    ARG_TABLE = [
        {'name': 'entity-id', 'help_text': 'Entity ID of the IdP'},
        {
            'name': 'ecp-endpoint-url',
            'help_text': 'ECP endpoint URL of the IdP'
        },
        {'name': 'username', 'help_text': 'Username to use on login to IdP'},
        {'name': 'password', 'help_text': 'Password to use on login to IdP'},
        {
            'name': 'role-arn',
            'help_text': 'The Role ARN to select. '
                         'If the IdP returns a single Role it is autoselected.'
        },
        {'name': 'factor', 'help_text': 'The Duo factor to use on login'},
        {'name': 'passcode', 'help_text': 'A Duo passcode'},
        {
            'name': 'refresh',
            'default': 0,
            'cli_type_name': 'integer',
            'help_text': 'How often in seconds to refresh the STS credentials'
        },
        {
            'name': 'verbose',
            'action': 'count',
            'default': 0,
            'cli_type_name': 'integer',
            'help_text': 'Display verbose output'
        },
        {
            'name': 'ask-password',
            'action': 'store_true',
            'default': False,
            'help_text': 'Force prompt for password'
        },
        {
            'name': 'force-refresh',
            'action': 'store_true',
            'default': False,
            'help_text': 'Forces a login attempt to the IdP using cookies'
        },
        {
            'name': 'duration',
            'default': None,
            'help_text': 'time in seconds that the token will last'
        },
        {
            'name': 'disable-refresh',
            'default': False,
            'cli_type_name': 'boolean',
            'help_text': 'Disables automatic refresh of tokens'
        },
    ]

    UPDATE = False

    def _run_main(self, args: Namespace, parsed_globals):
        main(args, self._session)
        return 0


class Logout(BasicCommand):
    NAME = 'logout'
    DESCRIPTION = ("Kills the process that is renewing the user's"
                   "credentials.")
    SYNOPSIS = ('aws logut')

    ARG_TABLE = [
        {
            'name': 'verbose',
            'action': 'count',
            'default': 0,
            'cli_type_name': 'integer',
            'help_text': 'Display verbose output'
        },
    ]

    UPDATE = False

    def _run_main(self, args: Namespace, parsed_globals):
        logout(args, self._session)
        return 0


class Configure(BasicCommand):
    NAME = 'login'
    DESCRIPTION = ('''
Configure LOGIN options. If this command is run with no arguments,
you will be prompted for configuration values such as your IdP's
entity ID and its ECP endpoint URL.  You can configure a named
profile using the --profile argument. If your config file does not
exist (the default location is ~/.aws-login/config), it will be
created for you. To keep an existing value, hit enter when prompted
for the value.  When you are prompted for information, the current
value will be dis- played in [brackets]. If the config item has
no value, it be displayed as [None].

=======================
Configuration Variables
=======================

The following configuration variables are supported in the config
file:

* **entity_id** - The entity ID of the IDP to use for authenitication
* **ecp_endpoint_url** - The ECP endpoint URL of the IDP to use for authn
''')
    SYNOPSIS = ('aws login configure')

    ARG_TABLE = [
        {
            'name': 'verbose',
            'action': 'count',
            'default': 0,
            'cli_type_name': 'integer',
            'help_text': 'Display verbose output'
        },
    ]

    UPDATE = False

    EXAMPLES = ('''
To create a new configuration::\n
\n
    $ aws login configure
    Entity ID [None]: urn:mace:incommon:idp.edu
    ECP Endpoint URL [None]: https://idp.edu/idp/profile/SAML2/SOAP/ECP\n
\n
To update just the entity ID::\n
\n
    $ aws login configure
    Entity ID [urn:mace:incommon:idp.edu]: urn:mace:uncommon:foo.com
    ECP Endpoint URL [https://idp.edu/idp/profile/SAML2/SOAP/ECP]:
''')

    def _run_main(self, args: Namespace, parsed_globals):
        configure(args, self._session)
        return 0

class AccountNames(BasicCommand):
    NAME = 'account_names'
    DESCRIPTION = ('''
Print account names and account IDs
=======================
Configuration Variables
=======================

''')
    SYNOPSIS = ('aws login account_names')

    ARG_TABLE = [
        {
            'name': 'verbose',
            'action': 'count',
            'default': 0,
            'cli_type_name': 'integer',
            'help_text': 'Display verbose output'
        },
        {
            'name': 'save',
            'action': 'store_true',
            'default': False,
            'cli_type_name': 'boolean',
            'help_text': 'Update accounts.ini file with names'
        },
    ]

    EXAMPLES = ('''
To create a new configuration::\n
\n
    $ aws login configure
    Entity ID [None]: urn:mace:incommon:idp.edu
    ECP Endpoint URL [None]: https://idp.edu/idp/profile/SAML2/SOAP/ECP\n
\n
To update just the entity ID::\n
\n
    $ aws login configure
    Entity ID [urn:mace:incommon:idp.edu]: urn:mace:uncommon:foo.com
    ECP Endpoint URL [https://idp.edu/idp/profile/SAML2/SOAP/ECP]:
''')

    def _run_main(self, args: Namespace, parsed_globals):

        if args.save:
            save_account_names(args, self._session)
        else:
            print_account_names(args, self._session)

        return 0



from copy import copy
from typing import List

from awscli_login.exceptions import InvalidFactor

from .base import ProfileBase


class Creds():

    def __init__(self, username: str = None, password: str = None,
                 factor: str = None, passcode: str = None,
                 keyring: str = None) -> None:
        self.username = username
        self.password = password
        self.factor = factor
        self.passcode = passcode
        self.keyring = keyring

    def get(self) -> List[str]:
        r = []

        if self.username is not None:
            r.append(self.username)

        if self.factor is not None:
            r.append(self.factor)

        if self.passcode is not None:
            r.append(self.passcode)

        return r


class GetCredsProfileBase(ProfileBase):

    def mock_get_credentials_inputs(self, inputs: Creds):
        mock_input = self.patcher('builtins.input', autospec=True,
                                  side_effect=inputs.get())

        mock_password = self.patcher(
            'awscli_login.config.getpass',
            autospec=True,
            side_effect=[inputs.password]
        )

        mock_get_password = self.patcher(
            'awscli_login.config.get_password',
            autospec=True,
            side_effect=[inputs.keyring]
        )

        mock_set_password = self.patcher(
            'awscli_login.config.set_password',
            autospec=True
        )

        return (
            mock_input,
            mock_password,
            mock_get_password,
            mock_set_password,
            inputs,
        )

    def assertGetCredentialsMocksCalled(
            self,
            mock_input,
            mock_password,
            mock_get_password,
            mock_set_password,
            inputs,
    ):
        # Make sure the user was prompted for exactly the inputs given
        self.assertEqual(mock_input.call_count, len(inputs.get()))

        if inputs.password:
            try:  # Python 3.6 only
                mock_password.assert_called_once()
            except AttributeError:
                self.assertEqual(
                    1,
                    mock_password.call_count
                )
        else:
            try:  # Python 3.6 only
                mock_password.assert_not_called()
            except AttributeError:
                self.assertEqual(
                    0,
                    mock_password.call_count
                )

        if self.profile.enable_keyring:
            try:  # Python 3.6 only
                mock_get_password.assert_called_once()
                mock_set_password.assert_called_once()
            except AttributeError:
                self.assertEqual(
                    1,
                    mock_get_password.call_count
                )
                self.assertEqual(
                    1,
                    mock_set_password.call_count
                )
        else:
            try:
                mock_get_password.assert_not_called()
                mock_set_password.assert_not_called()
            except AttributeError:
                self.assertEqual(
                    0,
                    mock_get_password.call_count
                )
                self.assertEqual(
                    0,
                    mock_set_password.call_count
                )

    def _test_get_credentials(self, outputs: Creds) -> None:
        # Ensure actual outputs equal expected outputs
        errors = []
        mesg = 'get_credentials returned %s: %s. Expected: %s.'
        self.outputs = outputs

        usr, pwd, hdr = self.profile.get_credentials()

        pairs = [
            ('username', usr),
            ('password', pwd),
            ('factor', hdr.get('X-Shiboleth-Duo-Factor')),
            ('passcode', hdr.get('X-Shiboleth-Duo-Passcode')),
        ]

        for name, ret in pairs:
            expected = getattr(self.outputs, name)

            if (ret != expected):
                errors.append(mesg % (name, ret, expected))

        if errors:
            raise AssertionError('\n'.join(errors))


class GetCredsMinProfileBase(GetCredsProfileBase):
    """ Minimal default profile with no tests. """

    def setUp(self) -> None:
        super().setUp()
        self.login_config = """
[default]
ecp_endpoint_url = foo
    """
        self.Profile()

    def test_get_credentials_bad_factor(self):
        """ Given a bad factor on login InvalidFactor should be raised. """
        inputs = Creds(username="user", password="secret", factor="bozo")

        mocks = self.mock_get_credentials_inputs(inputs)
        with self.assertRaises(InvalidFactor):
            usr, pwd, hdr = self.profile.get_credentials()
        self.assertGetCredentialsMocksCalled(*mocks)

    def test_get_credentials_prompt_for_all(self):
        """ Should prompt for all user credentials. """
        inputs = Creds(username="user", password="secret", factor="passcode",
                       passcode="1234")
        outputs = copy(inputs)

        mocks = self.mock_get_credentials_inputs(inputs)
        self._test_get_credentials(outputs)
        self.assertGetCredentialsMocksCalled(*mocks)

    def test_get_credentials_phone_factor(self):
        """ Should accept phone factor. """
        inputs = Creds(username="user", password="secret", factor="phone")
        outputs = copy(inputs)

        mocks = self.mock_get_credentials_inputs(inputs)
        self._test_get_credentials(outputs)
        self.assertGetCredentialsMocksCalled(*mocks)

    def test_get_credentials_no_factor_given(self):
        """ Should accept empty factor. """
        inputs = Creds(username="user", password="secret", factor="")
        outputs = copy(inputs)
        outputs.factor = None

        mocks = self.mock_get_credentials_inputs(inputs)
        self._test_get_credentials(outputs)
        self.assertGetCredentialsMocksCalled(*mocks)


class GetCredsMinProfileNoDuoTest(GetCredsProfileBase):
    """ Test minimal profile with duo disabled. """

    def test_get_credentials_no_duo(self):
        """ Should prompt for just username/password. """
        self.login_config = """
[default]
ecp_endpoint_url = foo
factor = off
    """
        self.Profile()

        inputs = Creds(username="user", password="secret")
        outputs = copy(inputs)

        mocks = self.mock_get_credentials_inputs(inputs)
        self._test_get_credentials(outputs)
        self.assertGetCredentialsMocksCalled(*mocks)

    def test_get_credentials_prompt_for_passcode(self):
        """ Should prompt for passcode with factor in profile. """
        self.login_config = """
[default]
ecp_endpoint_url = foo
factor = passcode
    """
        self.Profile()

        inputs = Creds(username="user", password="secret", passcode="1234")
        outputs = copy(inputs)
        outputs.factor = 'passcode'

        mocks = self.mock_get_credentials_inputs(inputs)
        self._test_get_credentials(outputs)
        self.assertGetCredentialsMocksCalled(*mocks)

    def test_get_credentials_no_prompt_for_username(self):
        """ Should prompt for password/factor with username in profile. """
        self.login_config = """
[default]
ecp_endpoint_url = foo
username = user
    """
        self.Profile()

        inputs = Creds(password="secret", factor="push")
        outputs = copy(inputs)
        outputs.username = "user"

        mocks = self.mock_get_credentials_inputs(inputs)
        self._test_get_credentials(outputs)
        self.assertGetCredentialsMocksCalled(*mocks)

    def test_get_credentials_password_only(self):
        """ Should prompt for password only with username/factor in profile """
        self.login_config = """
[default]
ecp_endpoint_url = foo
username = user1
factor = auto
    """
        self.Profile()

        inputs = Creds(password="secret")

        outputs = copy(inputs)
        outputs.username = "user1"
        outputs.factor = "auto"

        mocks = self.mock_get_credentials_inputs(inputs)
        self._test_get_credentials(outputs)
        self.assertGetCredentialsMocksCalled(*mocks)

    def test_get_credentials_keyring_test_no_prompt(self):
        """ Should not prompt with keyring/username/factor in profile """
        self.login_config = """
[default]
ecp_endpoint_url = foo
username = user1
factor = auto
enable_keyring = true
    """
        self.Profile()

        inputs = Creds(keyring="secret")

        outputs = copy(inputs)
        outputs.username = "user1"
        outputs.factor = "auto"
        outputs.password = "secret"

        mocks = self.mock_get_credentials_inputs(inputs)
        self._test_get_credentials(outputs)
        self.assertGetCredentialsMocksCalled(*mocks)



from setuptools import setup, find_packages

from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='co-awscli-login',
    version='0.1.0a8',  # TODO change this to a git tag for Drone
    description='Plugin for the AWS CLI that retrieves and rotates '
    'credentials using SAML ECP and STS.',
    long_description=long_description,
    url='https://github.com/techservicesillinois/awscli-login',
    author='David D. Riddle',
    author_email='ddriddle@illinois.edu',
    classifiers=[
#        'Development Status :: 4 - Beta',
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='Amazon AWS SAML login access keys',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    python_requires='>=3.5',
    install_requires=[
       'awscli',
       'boto3',
       'botocore',
       'daemoniker',
       'keyring',
       'lxml',
       'psutil',
       'requests',
    ],
    extras_require={
        'test': [
            'Sphinx',
            'coverage',
            'sphinx-autodoc-typehints',
            'sphinx_rtd_theme',
            'tblib',
            'tox',
            'tox-pyenv',
            'wurlitzer',
        ],
    },
    test_suite="tests",
    project_urls={
        'Bug Reports':
            'https://github.com/techservicesillinois/awscli-login/issues',
        'Source': 'https://github.com/techservicesillinois/awscli-login',
    },
)



import logging
import re

import lxml.etree as ET

from base64 import b64encode
from datetime import datetime
from http.cookiejar import LWPCookieJar
from typing import List, Tuple
from typing import cast
from uuid import uuid4

from lxml.etree import XMLSyntaxError
from lxml.etree import tostring, Element, SubElement
from requests import Session
from requests.cookies import RequestsCookieJar

from .exceptions import (
    AuthnFailed,
    InvalidSOAP,
    MissingCookieJar,
    RoleParseFail,
)
from .typing import Role, Headers
from .util import secure_touch

SAML_SUCCESS = "urn:oasis:names:tc:SAML:2.0:status:Success"

ns = {
       'S': 'http://schemas.xmlsoap.org/soap/envelope/',
       'saml2': 'urn:oasis:names:tc:SAML:2.0:assertion',
       'saml2p': 'urn:oasis:names:tc:SAML:2.0:protocol',
}

logger = logging.getLogger(__name__)


def utcnow() -> datetime:
    """ This is a wrapper function to ease testing with Mocks. """
    return datetime.utcnow()


def raise_if_saml_failed(soap: bytes) -> None:
    """
    Parses a SAML SOAP response to determine if authentication
    succeeded or failed.

    Args:
        soap: A byte string containing a SOAP response from an
        IdP.

    Raises:
        AuthnFailed: If a SAML success code was not returned.
        XMLSyntaxError: If the SOAP response contains syntax errors.
    """
    xml = ET.fromstring(soap)

    elem = xml.find('S:Body/saml2p:Response/saml2p:Status' +
                    '/saml2p:StatusCode' +
                    '[@Value="' + SAML_SUCCESS + '"]', ns)

    if elem is None:
        raise AuthnFailed


def saml_login(url: str, jar: LWPCookieJar,
               username: str = None, password: str = None,
               headers: Headers = None) -> bytes:
    """
    Generates and posts a SAML AuthNRequest to an IdP.

    Args:
        url: ECP endpoint URL for the IdP.
        username: Username to provide to the IdP.
        password: Password to provide to the IdP.
        headers: optional headers to provide to the IdP.

    Returns:
        The SOAP response from the IdP.
    """
    s = Session()
    s.cookies = cast(RequestsCookieJar, jar)
    s.headers.update({'Content-Type': 'text/xml', 'charset': 'utf-8'})

    envelope = authn_request()
    auth = (username, password) if username and password else None
    logger.debug("POST %s\nheaders: %s\npayload %s" %
                 (url, headers, envelope))
    r = s.post(url, data=envelope, headers=headers, auth=auth)
    logger.debug("POST returned: %s" % r.content)

    r.raise_for_status()
    try:
        raise_if_saml_failed(r.content)
    except XMLSyntaxError:
        raise InvalidSOAP(url)
    return r.content


def authenticate(url: str, cookies: str,
                 username: str, password: str,
                 headers: Headers) -> Tuple[str, List[Role]]:
    """
    Authenitcate with user credentials to IdP.

    Args:
        url: ECP endpoint URL for the IdP.
        cookies: A path to a cookie jar.
        username: Username to provide to the IdP.
        password: Password to provide to the IdP.
        headers: Optional headers to provide to the IdP.

    Returns:
        A base 64 encoded SAML assertion string, and a list of
        tuples containing a SAML provider ARN and a Role ARN.
    """
    jar = LWPCookieJar(cookies)
    soap = saml_login(url, jar, username, password, headers)

    mesg = "Successfully authenticated with username/password"
    logger.info(mesg + " to endpoint: " + url)

    secure_touch(cookies)
    jar.save(ignore_discard=True)
    logger.info("Saved cookies to jar: " + jar.filename)

    return parse_soap_response(soap)


def refresh(url: str, cookies: str) -> Tuple[str, List[Role]]:
    """
    Reauthenticate with cookies to IdP.

    Args:
        url: ECP endpoint URL for the IdP.
        cookies: A path to a cookie jar.

    Returns:
        A base 64 encoded SAML assertion string, and a list of
        tuples containing a SAML provider ARN and a role ARN.
    """
    jar = LWPCookieJar(cookies)
    try:
        jar.load(ignore_discard=True)
        logger.info("Loaded cookie jar: " + cookies)
    except FileNotFoundError:
        raise MissingCookieJar(url)

    soap = saml_login(url, jar)

    mesg = "Successfully authenticated with cookies"
    logger.info(mesg + " to endpoint: " + url)

    return parse_soap_response(soap)


def parse_soap_response(soap: bytes) -> Tuple[str, List[Role]]:
    """
    Parses SAML SOAP response for SAML assertion and SAML attributes:
    provider & role ARN(s).

    Args:
        soap: A SAML Response

    Returns:
        A base 64 encoded SAML assertion string, and a list of
        tuples containing a SAML provider ARN and a role ARN.
    """
    body = ET.fromstring(soap)
    resp = body.find('S:Body/saml2p:Response', ns)

    roles = resp.findall("saml2:Assertion/" +
                         "saml2:AttributeStatement/saml2:Attribute[@Name=" +
                         "'https://aws.amazon.com/SAML/Attributes/Role']/" +
                         "saml2:AttributeValue", ns)

    return b64encode(tostring(resp)).decode("us-ascii"), parse_role_arns(roles)


def parse_role_arns(roles: List[SubElement]) -> List[Role]:
    """
    Parses SAML Attributes for SAML provider and Role ARNS.

    Args:
        roles: List of XML SAML Attributes.

    Returns:
        A list of tuples containing a SAML provider ARN and a
        Role ARN.

    Raises:
        SAML: If unable to find a SAML provider or Role ARN.
    """
    role_arns = []
    role_regex = re.compile('.*(arn:aws:iam::([0-9]+):role/([^,:]+)).*')
    saml_regex = re.compile(
        '.*(arn:aws:iam::([0-9]+):saml-provider/([^,:]+)).*'
    )

    for role in roles:
        arn = role_regex.match(role.text)
        saml = saml_regex.match(role.text)

        if arn and saml:
            role_arns.append((saml.group(1), arn.group(1)))
        else:
            raise RoleParseFail(role.text)

    return role_arns


def authn_request() -> bytes:
    """
    Generates an ECP SAML AuthNRequest for the Amazon SP.

    Returns:
        A SOAP byte string.
    """
    now = utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
    uuid = '_' + str.upper(str(uuid4())).replace('-', '')

    SOAP = "{%s}" % ns['S']
    SAML2 = "{%s}" % ns['saml2']
    SAML2P = "{%s}" % ns['saml2p']

    envelope = Element(SOAP + "Envelope", nsmap=ns)
    body = SubElement(envelope, SOAP + "Body")

    attr = {
       'AssertionConsumerServiceURL': "https://signin.aws.amazon.com/saml",
       'ID': uuid,
       'IssueInstant': now,
       'ProtocolBinding': "urn:oasis:names:tc:SAML:2.0:bindings:PAOS",
       'Version': '2.0',
    }
    authn_request = SubElement(body, SAML2P + "AuthnRequest", **attr)

    issuer = SubElement(authn_request, SAML2 + "Issuer")
    issuer.text = "urn:amazon:webservices"

    return tostring(envelope)
